/* +-------------+---------+-----------------------+
   | John Breaux | jab0910 | JohnBreaux@my.unt.edu |
   +-------------+---------+-----------------------+
   |              Created  2022-04-17              |
   +-----------------------------------------------+  */
#include "read.hpp"

#include <fstream>
#include <regex>

#include "graph.hpp"

//* Vector of regex rules, in order of highest to lowest precedence
std::vector<std::regex> patterns = {
   //* Rule 1: A line containing % is a comment, and should not be parsed
   std::regex ("(%)"),
   //* Rule 2: A line containing `num_processes` contains the number of processes
   std::regex ("(num_processes)\\D*(\\d+)"),
   //* Rule 3: A line containing `num_resources` contains the number of resources
   std::regex ("(num_resources)\\D*(\\d+)"),
   //* Rule 4: A line containing integers should be saved somewhere
   std::regex ("(-?[0-9]+.*-?[0-9]+)") // Magical!
};

/* read:
*     Read from a file of name `filename`
   params:
      @param filename  Name of the file to read graph from
   returns:
      @returns int: error code representing the list of errors in the lowest 6 bits
*/
int graph::read (std::string filename) {
   // Open file with name filename as input fstream
   std::ifstream f (filename);
   if (!f) {
      print_error_message (INVALID_FILENAME, filename);
      return INVALID_FILENAME;
   } // File bad!
   // Read the entire file
   while (f && !f.eof ()) {
      // acquire a line
      std::string line;
      std::getline (f, line);
      // Iterate over each pattern, and grab the associated data
      for (const auto &pattern: patterns) {
         std::smatch res;
         // Capture data with a regex search, using the pattern
         if (std::regex_search (line, res, pattern)) {
            // get the pattern type, value
            std::string type = res.format ("$1"), value = res.format ("$2");
            // Handle the pattern
            //*  Rule 1: If line is a comment, ignore it and move on
            if (type == "%") break;

            //*  Rule 2: If num_processes= matched, assign value to num_processes
            else if (type == "num_processes") num_processes = std::stoi (value);

            //*  Rule 3: If num_resources= matched, assign value to num_resources
            else if (type == "num_resources") num_resources = std::stoi (value);

            //*  Rule 4: If this line is a comma-separated list of numbers:
            //*    If resource_counts is empty, this is the resource counts,
            //*    else it's part of the Matrix Reloaded
            else {
               if (resource_counts.size ()) m.push_back (stovi (type));
               else resource_counts = stovi (type);
            }
            break;
         }
      }
   }

   //* Test validity of the input file
   // Check if all sections are properly defined:
   int ret = 0;
   if (!num_processes)
      ret |= INVALID_PROCESSES;
   if (!num_resources)
      ret |= INVALID_RESOURCES;
   if (resource_counts.size () != num_resources)
      ret |= INVALID_COUNTS;
   if (m.size () != num_processes + num_resources)
      ret |= INVALID_ROWS;
   for (auto i: m) {
      if (i.size () != num_processes + num_resources) ret |= INVALID_COLUMNS;
   }
   // Print the appropriate error message
   if (ret) print_error_message (ret, filename);

   // fstream destructor closes file when going out of scope
   // in this house we RAII sometimes.
   return ret;
}

std::vector<int> stovi (const std::string &s) {
   // Create the number classification rule
   std::regex rule ("-?[0-9]+");
   // Make a new tokenizing iterator using the rule
   std::sregex_token_iterator start {s.begin (), s.end (), rule, 0}, end;
   // Match the numbers
   std::vector<int> res;
   for (auto i = start; i != end; ++i) res.push_back (atoi (i->str ().c_str ()));
   // give the numbers back
   return res;
}

void print_error_message (int error, std::string filename) {
   // Error message strings
   const char err_please[] = "Check to ensure your input file is correct and follows the Rules.\n";
   const char err_undefn[] = "%s: Definition failed for %s\n";
   const char err_nofile[] = "%s: File not found.\n";
   const char *fn          = filename.c_str ();
   // Test for each kind of error message, and print the appropriate messages
   if (error & graph::INVALID_PROCESSES)
      fprintf (stderr, err_undefn, fn, "num_processes");
   if (error & graph::INVALID_RESOURCES)
      fprintf (stderr, err_undefn, fn, "num_resources");
   if (error & graph::INVALID_COUNTS)
      fprintf (stderr, err_undefn, fn, "resource counts");
   if (error & graph::INVALID_ROWS)
      fprintf (stderr, err_undefn, fn, "adjacency matrix rows");
   if (error & graph::INVALID_COLUMNS)
      fprintf (stderr, err_undefn, fn, "adjacency matrix columns");
   if (error & graph::INVALID_FILENAME)
      fprintf (stderr, err_nofile, fn); // file not found
   if (error) printf (err_please);
}